"""Components for Python compatibility, introspection, and reflection.

Notes:
    The functionality here enables us to support multiple versions of Python to maintain
    forwards- and backwards-compatibility. This enables developers to leverage the
    bleeding edge of Python typing as new features and innovations are introduced,
    without having to upgrade to a new version of the language.

    Think of this package as a [polyfill][polyfill]{.external} for Python typing.

    [polyfill]: https://en.wikipedia.org/wiki/Polyfill_(programming)#:~:text=In%20software%20development%2C%20a%20polyfill,not%20natively%20support%20the%20feature.
"""
